import numpy as np
import scipy.optimize as spo
import isopy.dtypes as _dtypes
import itertools as _itertools
#import matplotlib.pyplot as plt
from matplotlib import cm, colors

__all__ = []

"""
Functions for doublespike data reduction
"""

class DoubleSpikeInversionOutput:
    def __init__(self, method, lambda_, alpha, beta, Fnat, Fins, spike_faction, sample_fraction, Q):
        self.method = method

        self.lambda_ = lambda_
        self.alpha = alpha
        self.beta = beta

        self.Fnat = Fnat
        self.Fins = Fins

        self.spike_fraction = spike_faction
        self.sample_fraction = sample_fraction
        self.Q = Q

    def delta(self, mass_ratio):
        #use Fnat to calcualte a offset for the mass ratio given
        pass

    def ppm(self, mass_ratio):
        pass


def ds_inversion_rudge(m, n, T, P, xtol = 1e-10):
    """
    Double spike inversion using the method of Rudge et al. 2009, Chemical Geology 265 420-431.

    Parameters
    ----------
    self
    m : ndarray
        Measured isotope ratios. Must be an 1-dim array of size 3 or a 2-dim array with first dimension of 3.
    n : ndarray
        Reference isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `m`.
    T : ndarray
        Spike isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `m`.
    P : ndarray
        Log of mass isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `m`.
    xtol : float
        Required tolerance for inversion

    Returns
    -------
    DoubleSpikeInversionOutput
    """
    mndim = m.ndim
    if m.size == 3: m = m.reshape(3,1)
    elif m.ndim == 1: raise ValueError()
    elif m.ndim == 2:
        if m.shape[0] != 3: raise ValueError()
    else: raise ValueError('m has to many dimensions ({})'.format(m.ndim))

    if n.size == 3: n = np.ones(m.shape) * n.reshape(3,1)
    elif n.shape != m.shape: raise ValueError('Shape of n {} does not match shape of m {}'.format(n.shape, m.shape))

    if T.size == 3: T = np.ones(m.shape) * T.reshape(3,1)
    elif T.shape != m.shape: ValueError('Shape of T {} does not match shape of m {}'.format(T.shape, m.shape))

    if P.size == 3: P = np.ones(m.shape) * P.reshape(3,1)
    elif P.shape != m.shape: ValueError('Shape of P {} does not match shape of m {}'.format(P.shape, m.shape))

    A = np.transpose([T-n, -n*P, m*P])
    b = np.transpose(m-n)
    x0 = np.transpose(np.linalg.solve(A,b))

    x, infodict, ier, mesg =  spo.fsolve(_inversion_rudge_function, x0, (P, n, T, m), None, True, xtol = xtol)

    if ier != 1:
        x = np.ones(m.shape) * np.nan
    error_message = mesg #Shoudl this raises an errror. I guess it primaraly happens when there are o many htings to solve

    if mndim == 2: x = x.reshape(3,-1)
    lambda_ = x[0]
    alpha = x[1]/(1-lambda_)
    beta = x[2]

    Fnat = alpha * -1
    Fins = beta

    spike_fraction = (1+((1-lambda_)/lambda_)*((1+np.sum((n/np.exp(alpha*P)), axis = 0))/(1+np.sum(T, axis = 0))))**-1
    sample_fraction = (1-spike_fraction)
    Q = sample_fraction / spike_fraction

    return DoubleSpikeInversionOutput('Rudge', lambda_, alpha, beta, Fnat, Fins, spike_fraction, sample_fraction, Q)

def _inversion_rudge_function( x, P, n, T, m):
    x  = x.reshape(3,-1)
    lambda_ = x[0]
    alpha = x[1]/(1-lambda_)
    beta = x[2]

    return ((lambda_ * T) + ((1-lambda_) * n * np.exp(-alpha * P)) - m * np.exp(-beta * P)).flatten()

def _inversion_rudge_jacobian( x, P, n, T, m):
    #follwing code only works for size = 3
    x = x.reshape(3,-1)
    lambda_ = x[0]
    alpha = x[1]/(1-lambda_)
    beta = x[2]
    output = np.transpose([T - (n * np.exp(-alpha * P) * (1 + alpha * P)), -n * P * np.exp(-alpha * P), m * P * np.exp(-beta * P)])
    return output

def ds_inversion_siebert(MS, ST, SP, Mass, Fnat_guess = 2, Fins_guess = -0.00001, loop_one_iterations = 3, loop_two_iterations=6):
    """
    Double spike inversion using the method of Siebert et al. 2001, Geochemistry, Geophysics, Geosystems 2,

    Parameters
    ----------
    self
    MS : ndarray
        Measured isotope ratios. Must be an 1-dim array of size 3 or a 2-dim array with first dimension of 3.
    ST : ndarray
        Reference isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `MS`.
    SP : ndarray
        Spike isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `MS`.
    Mass : ndarray
        Log of mass isotope ratios. Must be a 1-dim array of size 3 or a 2-dim array with the same shape as `MS`.

    Fnat_guess : float
        Starting guess for Fnat
    Fins_guess : float
        Starting guess for Fins
    loop_one_iterations : int
        Number of iterations of the outer loop
    loop_two_iterations
        Number of iterations of the inner loop

    Returns
    -------
    DoubleSpikeInversionOutput
    """
    output_index = 0
    mndim = MS.ndim
    if MS.size == 3:
        MS = MS.reshape(3, 1)
    elif MS.ndim == 1: raise ValueError()
    elif MS.ndim == 2:
        if MS.shape[0] != 3: raise ValueError()
    else: raise ValueError('MS has to many dimensions ({})'.format(MS.ndim))

    if ST.size == 3: ST = ST.reshape(3)
    elif ST.shape != MS.shape: raise ValueError('Shape of ST {} does not match shape of MS {}'.format(ST.shape, MS.shape))

    if ST.size == 3: ST = ST.reshape(3)
    elif ST.shape != MS.shape: ValueError('Shape of ST {} does not match shape of MS {}'.format(ST.shape, MS.shape))

    if Mass.size == 3: Mass = Mass.reshape(3)
    elif Mass.shape != MS.shape: ValueError('Shape of Mass {} does not match shape of MS {}'.format(Mass.shape, MS.shape))

    dlen = MS.shape[1]
    SA = np.zeros((loop_one_iterations, 3, dlen), dtype = np.float64)
    Fins = np.zeros((loop_one_iterations, 3, dlen), dtype = np.float64)
    Fnat = np.zeros((loop_one_iterations, 3, dlen), dtype = np.float64)
    MT = np.zeros((3, dlen), dtype = np.float64)

    x = 0
    y = 1
    z = 2
    Ri = output_index

    Fnat_Ri = Fnat_guess
    Fins_Ri = Fins_guess

    Q = np.zeros(dlen, dtype = np.float64)

    for i1 in range(loop_one_iterations):
        SA[i1, x,:] = ST[x] * Mass[x] ** Fnat_Ri
        SA[i1, y,:] = ST[y] * Mass[y] ** Fnat_Ri
        SA[i1, z,:] = ST[z] * Mass[z] ** Fnat_Ri

        a = (ST[y] * (SA[i1, z] - SP[z]) + SA[i1, y] * (SP[z] - ST[z]) + SP[y] * (ST[z] - SA[i1, z])) / (ST[y] * (SA[i1, x] - SP[x]) + SA[i1, y] * (SP[x] - ST[x]) + SP[y] * (ST[x] - SA[i1, x]))
        b = (ST[x] * (SA[i1, z] - SP[z]) + SA[i1, x] * (SP[z] - ST[z]) + SP[x] * (ST[z] - SA[i1, z])) / (ST[x] * (SA[i1, y] - SP[y]) + SA[i1, x] * (SP[y] - ST[y]) + SP[x] * (ST[y] - SA[i1, y]))
        c = ST[z] - a * ST[x] - b * ST[y]

        for i2 in range(loop_two_iterations):
            MT[x] = MS[x] * Mass[x] ** -Fins_Ri
            MT[y] = MS[y] * Mass[y] ** -Fins_Ri
            MT[z] = MS[z] * Mass[z] ** -Fins_Ri

            d = (MS[z] - MT[z]) / (MS[x] - MT[x])
            e = MS[z] - d * MS[x]
            f = (MS[z] - MT[z]) / (MS[y] - MT[y])
            g = MS[z] - f * MS[y]

            MT[x] = (b * g - b * e + e * f - c * f) / (a * f + b * d - d * f)
            MT[y] = (a * e - a * g + d * g - c * d) / (a * f + b * d - d * f)
            MT[z] = a * MT[x] + b * MT[y] + c

            #gives positive Fins
            Fins[i1, 0] = np.log(MS[x] / MT[x]) / np.log(Mass[x])
            Fins[i1, 1] = np.log(MS[y] / MT[y]) / np.log(Mass[y])
            Fins[i1, 2] = np.log(MS[z] / MT[z]) / np.log(Mass[z])
            Fins_Ri = Fins[i1, Ri]

        a = (MS[y] * (MT[z] - SP[z]) + MT[y] * (SP[z] - MS[z]) + SP[y] * (MS[z] - MT[z])) / (MS[y] * (MT[x] - SP[x]) + MT[y] * (SP[x] - MS[x]) + SP[y] * (MS[x] - MT[x]))
        b = (MS[x] * (MT[z] - SP[z]) + MT[x] * (SP[z] - MS[z]) + SP[x] * (MS[z] - MT[z])) / (MS[x] * (MT[y] - SP[y]) + MT[x] * (SP[y] - MS[y]) + SP[x] * (MS[y] - MT[y]))
        c = MS[z] - a * MS[x] - b * MS[y]

        d = (ST[z] - SA[i1, z]) / (ST[x] - SA[i1, x])
        e = ST[z] - d * ST[x]
        f = (ST[z] - SA[i1, z]) / (ST[y] - SA[i1, y])
        g = ST[z] - f * ST[y]

        SA[i1, x] = (b * g - b * e + e * f - c * f) / (a * f + b * d - d * f)
        SA[i1, y] = (a * e - a * g + d * g - c * d) / (a * f + b * d - d * f)
        SA[i1, z] = a * SA[i1, x] + b * SA[i1, y] + c

        Fnat[i1, x] = np.log(SA[i1, x] / ST[x]) / np.log(Mass[x])
        Fnat[i1, y] = np.log(SA[i1, y] / ST[y]) / np.log(Mass[y])
        Fnat[i1, z] = np.log(SA[i1, z] / ST[z]) / np.log(Mass[z])
        Fnat_Ri = Fnat[i1, Ri]


    sample_fraction = ((1 / (MT[x] + MT[y] + MT[z] + 1)) - (1 / (SP[x] + SP[y] + SP[z] + 1))) / ((1 / (SA[loop_one_iterations - 1, x] + SA[loop_one_iterations - 1, y] + SA[loop_one_iterations - 1, z] + 1)) - (1 / (SP[x] + SP[y] + SP[z] + 1)))
    Fnat_f = Fnat[-1, Ri]
    Fins_f = Fins[-1, Ri]

    lambda_ = ((1-sample_fraction) ** (-1) - 1) / ((1 + np.sum(SA[-1], axis=0)) / (1 + np.sum(SP, axis=0)))
    lambda_ = 1 - (lambda_ / (lambda_ + 1))

    if mndim == 1:
        sample_fraction =  sample_fraction[0]
        Fnat_f= Fnat_f[0]
        Fins_f = Fins_f[0]
        lambda_ = lambda_[0]

    spike_fraction = (1-sample_fraction)
    Q = sample_fraction/spike_fraction

    alpha = Fnat_f * -1
    beta = Fins_f

    return DoubleSpikeInversionOutput(
        'Siebert', lambda_, alpha, beta, Fnat_f, Fins_f, spike_fraction, sample_fraction, Q)

def ds_inversion(measured, standard, spike, mass, inversion_ratios = None, method ='rudge', **method_kwargs):
    """
    Double spike inversion

    Examples
    --------
    lots of usage examples here

    Parameters
    ----------
    measured : RatioArray
        Measured isotope ratios
    standard : RatioArray, IsopyDict
        References isotope ratios or a dict or references values
    spike : RatioArray, IsopyDict
        Spike isotope ratios or a dict or references values
    mass : RatioArray, IsopyDict
        Mass isotope ratios or a dict or references values
    inversion_ratios : RatioList, None
        List of keys in measured that will be used for the inversion. Only necessary if measured contains more than 3 keys.
    method : str
        Inversion method to be used. Options are 'Rudge' and 'Siebert'.
    method_kwargs
        Keyword arguments for inversion method. See `inversion_rudge` and `inversion_siebert` for list of possible arguments.

    Returns
    -------
    DoubleSpikeInversionOutput
    """
    #Measured must be RatioArray
    #Standard, spike, and mass can be either RatioArray or dict
    if inversion_ratios is None:
        if len(measured.keys()) > 3: raise ValueError('If "measured" has more than 3 keys then "inversion_ratios" must be given')
        inversion_ratios = measured.keys()
    m = np.array([measured[x] for x in inversion_ratios])
    n = np.array([standard[x] for x in inversion_ratios])
    T = np.array([spike[x] for x in inversion_ratios])
    P = np.array([mass[x] for x in inversion_ratios])

    if method == 'rudge': return ds_inversion_rudge(m, n, T, np.log(P), **method_kwargs)
    elif method == 'siebert': return ds_inversion_siebert(m, n, T, P)
    else: raise ValueError('method "{}" not recognized'.format(method), **method_kwargs)


def johnson_nyquist_noise(voltages, resistors = None, integration_time = None, include_counting_statistics = True,
                          k = None, T = None, R = None, cpv = None):

    #Follows the equations from Liu & Pearson, Chemical Geology, 2014, 10, p301-311
    if integration_time is None: integration_time = np.float64(8.389)
    if k is None: k = np.float64(1.3806488E-023)
    #e = np.float64(1.602176565E-019)
    if T is None: T = np.float64(309)
    if R is None: R = np.float64(1E11)
    if cpv is None: cpv = np.float64(62500 * 1000)

    voltages = np.asarray(voltages)
    if resistors is None: resistors = [R] * voltages.shape[-1]
    resistors = np.asarray(resistors)

    t_noise = np.sqrt((4 * k * T * resistors) / integration_time) / (voltages * (resistors / R))

    if include_counting_statistics:
        c_stat = np.sqrt(1 / (voltages * cpv * integration_time))
        return np.sqrt(c_stat**2 + t_noise **2)*voltages
    else:
        return t_noise

def ds_uncertainty_grid(standard, spike1, spike2, mass, *, isotopes = None, max_voltage = 10, resistors = None,
                        integration_time = 4, cycles = 100, resolution = 19, Fnat = 0.1, Fins = 1.6,
                        ds_method ='Siebert', fast_mode = True, output_attr = 'alpha'):

    if isinstance(standard, _dtypes.IsotopeArray):
        if isotopes is None: isotopes = standard.keys()
        std = np.array([standard[i] for i in isotopes]).flatten()
    elif isinstance(standard, _dtypes.IsopyDict):
        std = np.array([standard[i] for i in isotopes]).flatten()
    else:
        std = np.asarray(standard).flatten()

    if std.size != 4: raise ValueError('standard must contain exactly four values')

    if isinstance(spike1, str):
        if isotopes is None: raise ValueError('spike1: "isotopes" not set')
        sp1 = np.zeros(4, dtype='f8')
        sp1[isotopes.index(spike1)] = 1
    elif isinstance(spike1, _dtypes.IsotopeArray):
        if isotopes is None: raise ValueError('spike1: "isotopes" not set')
        std = np.array([spike1[i] for i in isotopes]).flatten()
    else:
        sp1 = np.asarray(spike1).flatten()
    if sp1.size != 4: raise ValueError('spike1 must contain exactly four values')

    if isinstance(spike2, str):
        if isotopes is None: raise ValueError('spike2: "isotopes" not set')
        sp2 = np.zeros(4, dtype='f8')
        sp2[isotopes.index(spike2)] = 1
    elif isinstance(spike2, _dtypes.IsotopeArray):
        if isotopes is None: raise ValueError('spike2: "isotopes" not set')
        std = np.array([spike2[i] for i in isotopes]).flatten()
    else:
        sp2 = np.asarray(spike2).flatten()
    if sp2.size != 4: raise ValueError('spike2 must contain exactly four values')

    if isinstance(mass, (_dtypes.IsotopeArray, _dtypes.IsopyDict)):
        if isotopes is None: raise ValueError('mass: "isotopes" not set')
        mas = np.array([mass[i] for i in isotopes]).flatten()
    else:
        mas = np.asarray(mass).flatten()
    if mas.size != 4: raise ValueError('mass must contain exactly four values')

    #At this point standard, spike1 and spike2 are arrays of 4.
    standard = std / np.sum(std)
    spike1 = sp1 / np.sum(sp1)
    spike2 = sp2 / np.sum(sp2)
    mass = mas

    prop = np.linspace(0.0,1, resolution+2)[1:-1]

    denom = np.argmax(spike2)
    rat_index = [True] * 4
    rat_index[denom] = False

    mass_ratio = mass[rat_index] / mass[denom]

    if ds_method == 'Rudge':
        method = Rudge_inversion
        mass_ratio2 = np.log(mass_ratio)

    elif ds_method == 'Siebert':
        method = Siebert_inversion
        mass_ratio2 = mass_ratio

    std_ratio = (standard[rat_index] / standard[denom]) * mass_ratio ** Fnat

    if fast_mode is False:
        z = np.zeros((resolution, resolution, cycles), dtype=np.float64) * np.nan
        for y in range(resolution):
            spike = prop[y] * spike1 + (1-prop[y]) * spike2
            sp_rat = spike[rat_index] / spike[denom]
            for x in range(resolution):
                sample = prop[x] * spike + (1 - prop[x]) * standard
                sample = sample * max_voltage/np.max(sample)
                noise = johnson_nyquist_noise(sample, resistors, integration_time)
                noisy_sample = np.random.normal(sample, noise, (cycles,4)).transpose()
                smp_rat = noisy_sample[rat_index] / noisy_sample[denom]
                smp_rat = np.transpose(smp_rat.transpose() * mass_ratio ** Fins)

                solution = method(smp_rat, std_ratio, sp_rat, mass_ratio2)
                z[y,x, :] = solution.__getattribute__(output_attr)

    if fast_mode is True:
        big = np.ones((resolution, resolution, cycles, 4))
        small = np.ones((resolution, resolution, 4))

        spike = big * spike1 * prop.reshape(resolution, 1, 1, 1) + big * spike2 * (1 - prop).reshape(resolution, 1, 1,                                                                                          1)
        smp = spike[:, :, 1, :] * prop.reshape(1, resolution, 1) + small * standard * (1 - prop).reshape(1, resolution,
                                                                                                         1)
        smp = smp.reshape(-1, 4)
        smp = smp * (max_voltage / np.max(smp, axis=1).reshape(-1, 1))
        noise = johnson_nyquist_noise(smp, resistors, integration_time)

        noisy_sample = np.random.normal(smp, noise, (cycles, resolution ** 2, 4))
        noisy_sample = noisy_sample.transpose().reshape(4, -1)
        smp_ratio = noisy_sample[rat_index] / noisy_sample[denom] * (mass_ratio ** Fins).reshape(3, 1)

        spike = spike.flatten().reshape(-1, 4).transpose()
        sp_ratio = spike[rat_index] / spike[denom]

        #TODO check error that comes from rudge and do a specieal error
        solution = method(smp_ratio, std_ratio, sp_ratio, mass_ratio2)
        z = solution.__getattribute__(output_attr).transpose().flatten().reshape(resolution, resolution, -1)

    std = np.nanstd(z, axis=2) * 2
    count = np.count_nonzero(~np.isnan(z), axis=2)

    std[count < cycles * 0.9] = np.nan

    if output_attr is 'alpha':
        mean = np.nanmean(z, axis=2)
        std[np.abs(mean - Fnat) > std] = np.nan

    argmin_z = np.nanargmin(std)
    min_x_pos = argmin_z % std.shape[0]
    min_y_pos = int(argmin_z / std.shape[0])

    #x = sp in sp+smp, y = sp1 in sp1+sp2
    return prop, prop, std / np.sqrt(count), (min_y_pos, min_x_pos)

def plot_ds_uncertainty_grid(*args, **kwargs):
    x, y, z, best = ds_uncertainty_grid(*args, **kwargs)

    levels = [0] + [x * 0.0001 for x in range(1, 10)] + [x * 0.001 for x in range(1, 10)] + [x * 0.01 for x in
                                                                                             range(1, 10)] + [0.1]
    min_z = z[best]
    min_x = x[best[1]]
    min_y = y[best[0]]

    title = 'Smallest uncertianty: {:.5f} ($2\\sigma/\\sqrt{}n{}$) ' \
            'at x: {:.2f}, y: {:.2f}'.format(min_z, '{', '}', min_x, min_y)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    fig.subplots_adjust(hspace=0.3)
    CS = ax.contourf(x, y, z, levels=levels, cmap=cm.get_cmap(cm.jet), norm=colors.PowerNorm(0.23))
    fig.colorbar(CS, ax=ax, ticks=[0, 0.001, 0.01, 0.1, 1])
    plt.xlabel('Spike fraction in Spike/Sample mix')
    plt.ylabel('Spike1 fraction in Spike1/Spike2 mix')
    plt.title(title)
    plt.show()


def ds_cocktail_list(standard, mass, spikes = None, isotopes = None, output_mass_ratio = None, output_multiple = 1000, **kwargs):
    if isinstance(standard, _dtypes.IsotopeArray):
        if isotopes is None: isotopes = standard.keys()
    elif isinstance(standard, _dtypes.IsopyDict):
        if isotopes is None: raise ValueError('"isotopes" not given')
    else:
        raise ValueError('"standard" must be either a IsotopeArrray or a IsopyDict')

    if isotopes is None: raise ValueError('"isotopes" not given')
    if len(isotopes) < 4: raise ValueError('At least 4 isotopes are needed')

    if spikes is None: spikes = {}
    if not isinstance(spikes, dict): raise ValueError('"spikes" must be a dict')
    for s in spikes:
        if not isinstance(s, _dtypes.IsotopeArray, _dtypes.IsopyDict):
            raise ValueError('Each entry in "spikes" must be either a IsotopeArrray or a IsopyDict')

    if not isinstance(mass, (_dtypes.IsotopeArray, _dtypes.IsopyDict)):
        raise ValueError('"mass" must be either a IsotopeArray or a IsopyDict')

    output = []
    for iso in _itertools.combinations(isotopes, 4):
        for spike in _itertools.combinations(iso, 2):
            x, y, z, best = uncertianty_grid(standard, spikes.get(spike[0], spike[0]), spikes.get(spike[1], spike[1]), mass, isotopes = iso, **kwargs)

            x_min = x[best[1]]
            y_min = y[best[0]]

            z_min = z[best]

            if output_mass_ratio is not None:
                z_min = (output_mass_ratio ** z[best] -1) * output_multiple

            output.append([tuple(iso), tuple(spike), (x_min, y_min, z_min), x, y, z, best])
    return output










import numpy as _np
from numpy.lib.function_base import array_function_dispatch


__all__ = ['sd', 'se', 'mad', 'nansd', 'nanse', 'nanmad']

#These functions will all call __array_function__

def _sd_dispatcher(a, axis=None):
    return (a,)

@array_function_dispatch(_sd_dispatcher)
def sd(a, axis=None):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.


    Shortcut for ``np.std(a, ddof=1, axis=axis)``.

    See documentation for `np.std <https://numpy.org/devdocs/reference/generated/numpy.std.html#numpy.std>`_
    for additional information.

    See Also
    --------
    :func:`nansd`
    """
    return _np.std(a, ddof = 1, axis=axis)

@array_function_dispatch(_sd_dispatcher)
def nansd(a, axis=None):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom, while ignoring NaNs.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.


    Shortcut for ``np.nanstd(a, ddof=1, axis=axis)``.

    See documentation for `numpy.nanstd <https://numpy.org/devdocs/reference/generated/numpy.nanstd.html#numpy.nanstd>`_
    for additional information.

    See Also
    --------
    :func:`sd`
    """
    return _np.nanstd(a, ddof = 1, axis=axis)


def _se_dispatcher(a, axis=None):
    return (a,)


@array_function_dispatch(_se_dispatcher)
def se(a, axis=None):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.

    See Also
    --------
    :func:`nanse`
    """
    if axis is None:
        n = a.size
        return _np.std(a, ddof=1) / _np.sqrt(n)
    else:
        n = a.shape[axis]
        return _np.std(a, ddof=1, axis=axis) / _np.sqrt(n)


@array_function_dispatch(_se_dispatcher)
def nanse(a, axis=None):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom, while ignoring NaNs.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.

    See Also
    --------
    :func:`se`
    """
    if axis is None:
        n = _np.count_nonzero(~_np.isnan(a))
        return  _np.nanstd(a, ddof=1) / _np.sqrt(n)
    else:
        n = _np.count_nonzero(~_np.isnan(a), axis=axis)
        return _np.nanstd(a, ddof=1, axis=axis) / _np.sqrt(n)


def _mad_dispatcher(a, axis=None, scale=None):
    return (a,)


@array_function_dispatch(_mad_dispatcher)
def mad(a, axis=None, scale = 1.4826):
    """
    Compute the median absolute deviation of the data along the given axis.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.
    scale : int, optional
        The scaling factor applied to the MAD. The default scale (1.4826) ensures consistency with the standard
        deviation for normally distributed data.


    Shortcut for
    ``scipy.stats.median_absolute_deviation(a, axis=axis, center=np.median, scale=scale, nan_policy='propagate')``.

    See documentation for `scipy.stats.median_absolute_deviation
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_absolute_deviation.html>`_
    for additional information.

    See Also
    --------
    :func:`nanmad`
    """
    if axis is None:
        med = _np.nanmedian(a)
        mad = _np.median(_np.abs(a - med))
    else:
        med = _np.apply_over_axes(_np.nanmedian, a, axis)
        mad = _np.median(_np.abs(a - med), axis=axis)

    return mad * scale


@array_function_dispatch(_mad_dispatcher)
def nanmad(a, axis=None, scale = 1.4826):
    """
    Compute the median absolute deviation along the specified axis, while ignoring NaNs.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int or None, optional
        Axis along which the range is computed. Default is ``None``. If ``None`` compute
        over the entire array.
    scale : int, optional
        The scaling factor applied to the MAD. The default scale (1.4826) ensures consistency with the standard
        deviation for normally distributed data.

    See Also
    --------
    :func:`nanmad`
    """

    if axis is None:
        med = _np.nanmedian(a)
        mad = _np.nanmedian(_np.abs(a - med))
    else:
        med = _np.apply_over_axes(_np.nanmedian, a, axis)
        mad = _np.nanmedian(_np.abs(a - med), axis=axis)

    return mad * scale






import numpy as _np
from isopy import dtypes as _dt
from isopy import exceptions as _e
from isopy.toolbox import general as _g
from isopy.toolbox import misc as _m

__all__ = ['make_sample', 'remove_mass_fractionation', 'add_mass_fractionation',
           'calculate_mass_fractionation_factor', 'mass_independent_correction',
           'add_isobaric_interferences', 'remove_isobaric_interferences']

"""
Functions for isotope data reduction
"""

def make_sample(element, interference_isotopes = None, interference_values= None, integrations = 100,
                integration_time=4, maxv = 10, fins=None, fnat=None, sprocess=None, std_abu=None, std_mass = None,
                sprocess_fractions = None):
    """
    Returns a simulated measurement for all isotopes of *element*.

    The measured values are calculated using a normal distribution of the Johnson-Nyquist noise and counting statistics.

    The sample is created in the following order:

    * s-process contributions added

    * natural fractionation is added

    * isobaric interferences are added

    * instrumental fractionation is added

    * Johnson-Nyquist noise is added

    Parameters
    ----------
    element : ElementString
        All isotopes of this element will be included in the returned array.
    interference_isotopes : IsotopeList, optional
        If given then each isotope in *interference_isotopes* is added to the returned array and the isobaric
        interferences of the element is added to all isotopes in the array.
    interference_values : list of floats, optional
        The value of each isotope in *interference_isotope* when isobaric interferences is calculated. Must be given
        if *interference_isotopes* is given.
    integrations : int, optional
        The number of integrations in the returned array. Default value is ``100``
    integration_time : float, optional
        The integration time for each integration in the returned array. Used to calculate the Johnson_Nyquist noise.
        Default value is ``4.0``
    maxv : float, optional
        The voltage of the *element* isotopes with the highest abundance. Default value is ``10.0``
    fins : float, optional
        If given then an instrumental fractionation factor of *fins* is added to the data.
    fnat : float, optional
        If given then an natural fractionation factor of *fnat* is added to the data.
    sprocess : float, optional
        If given then *sprocess* amounts of s-process isotopes are added to the data. Given in permil.
    std_abu : IsotopeArray, dict, optional
        An object with a ".get(item, default_value)" method containing the abundance of each isotope/ratio needed for
        this function. Default is the "best isotope fraction" reference values.
    std_mass : IsotopeArray, dict, optional
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values.
    sprocess_fractions : IsotopeArray, dict, optional
        An object with a ".get(item, default_value)" method containing the s-process fraction of each isotope needed
        for this function. Default is the "sprocess isotope fraction" reference values.

    Returns
    -------
    IsotopeArray
        An array with size of *integrations*
    """
    #Check input
    element = _e.check_type('element', element, _dt.ElementString, coerce=True)
    interference_isotopes = _e.check_type('interference_isotopes', interference_isotopes, _dt.IsotopeList,
                                          coerce=True, allow_none=True)
    interference_values = _e.check_type_list('interference_values', interference_values, _np.float, coerce=True,
                                             make_list=True, allow_none=(interference_isotopes is None))
    if interference_isotopes is not None and len(interference_values) != len(interference_isotopes):
        raise ValueError('size of "interference_isotopes" and "interference_values" do not match')
    integrations = _e.check_type('integrations', integrations, int, coerce=True)
    integration_time = _e.check_type('integration_time', integration_time, _np.float, float, coerce=True)
    maxv = _e.check_type('maxv', maxv, _np.float, float, coerce=True)
    fins = _e.check_type('fins', fins, _np.float, float, coerce=True)
    fnat = _e.check_type('fnat', fnat, _np.float, float, coerce=True)
    sprocess = _e.check_type('sprocess', sprocess, _np.float, float, coerce=True)
    std_abu = _e.check_reference_value('std_abu', std_abu, 'best isotope fraction')
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')
    sprocess_fractions = _e.check_reference_value('sprocess_fractions', sprocess_fractions,
                                                          'sprocess isotope fraction')

    data = std_abu.get(std_abu.isotope_keys().copy(element_symbol=element))
    denom = _np.argmax(data, axis=None)

    #add s-process anomaly
    if sprocess is not None: data = data + (data * sprocess_fractions.get(data.keys(), 0) * (sprocess / 1000))

    # add natural fractionation
    rat = data.ratio(denom)
    if fnat is not None: rat = add_mass_fractionation(rat, fnat)

    #Add isobaric interferences
    if interference_isotopes is not None:
        #add the interfering isotopes to the array
        #This is done here so that any isobaric interferences on these isotopes is included
        rat = _np.append(rat, _dt.RatioArray(rat.nrows, keys=[iso / denom for iso in interference_isotopes]))
        for i in range(len(interference_isotopes)):
            isotope = interference_isotopes[i]
            value = interference_values[i] / maxv
            rat = add_isobaric_interferences(rat, isotope, value,std_abu=std_abu, std_mass=std_mass)


    #add intrumental fractionation
    if fins is not None: rat = add_mass_fractionation(rat, fins)

    #convert back into isotopes and set max voltage
    data = rat.deratio(maxv)

    #calculate jk noise
    noise = _m.johnson_nyquist_noise(data, integration_time=integration_time)

    #create measured data
    random_generator = _np.random.default_rng()
    measured_data = _dt.IsotopeArray(integrations, keys=data.keys())
    for key in measured_data.keys():
        measured_data[key] = random_generator.normal(data[key], noise[key], integrations)

    return measured_data


def mass_independent_correction(data, mf_ratio, normalisation_factor=None, std_abu=None, std_mass=None):
    """
    A quick function for mass-independent data correction.

    The data is corrected for mass fractionation, isobaric interferences and finally, if *normalisation_factor* is given,
    normalised to the values in *std_abu*. If *normalisation_factor* is not given the unnormalised corrected data is
    returned.

    An interference correction will be applied for all isotopes that are different from the *mf_ratio* numerator
    element. This will be done together with the mass fractionation correction to account for isobaric interferences
    on the *mf_ratio*.

    Parameters
    ----------
    data : IsotopeArray
        The data to be corrected.
    mf_ratio : RatioString
        The data will be internally normalised to this ratio.
    normalisation_factor : float, optional
        If given the corrected data is passed to :func:`normalise_data` together with the *std_abu* and with
        *normalisation_factor*.
    std_abu : dict
        An object with a ".get(item, default_value)" method containing the abundance of each isotope/ratio needed for
        this function. Default is the "best isotope fraction" reference values.
    std_mass : dict
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values.

    Returns
    -------
    RatioArray
        The data corrected for mass-independent variation. Only the isotopes of the element in *mf_ratio* is returned.

    """
    data = _e.check_type('data', data, _dt.IsotopeArray, coerce=True)
    mf_ratio = _e.check_type('mf_ratio', mf_ratio, _dt.RatioString, coerce=True)
    normalisation_factor = _e.check_type('normalisation_factor', normalisation_factor, _np.float, str, coerce=True, allow_none=True)
    std_abu = _e.check_reference_value('std_abu', std_abu, 'best isotope fraction')
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    #Find the isotopes that can cause isobaric interferences.
    interference_isotopes = data.keys().copy(element_symbol_not=mf_ratio.numerator.element_symbol)

    #Convert the data into a ratio array
    rat = data.ratio(mf_ratio.denominator)

    #Do a combined mass fractionation and isobaric interference correction.
    #This can account for isobaric interferences on isotopes in *mf_ratio*
    prev_beta = 0
    for i in range(10):
        #Calculate the mass fractionation.
        beta = calculate_mass_fractionation_factor(rat2, mf_ratio, std_abu, std_mass)

        rat2 = rat.copy() #get a fresh copy of ratio for interference corrections.
        for infiso in interference_isotopes:
            rat2 = remove_isobaric_interferences(rat2, infiso, beta, std_abu, std_mass)

        if beta/prev_beta < 0.0001:
            break #Beta value has converged so no need for more iterations.

    #Remove the isotopes on interfering elements
    rat = rat2.copy(element_symbol= mf_ratio.numerator.element_symbol)

    if normalisation_factor is not None:
        #Normalise the corrected data relative to *std_abu* and return
        rat =  _g.normalise_data(rat, std_abu, normalisation_factor)

    # Return the corrected data
    return rat


def calculate_mass_fractionation_factor(data, ratio, std_abu=None, std_mass=None):
    """
    Calculate the mass fractionation factor for a given ratio in *data*.

    .. math::
        \\alpha = \\ln{( \\frac{r_{n,d}}{R_{n,d}} )} * \\frac{1}{ \\ln{( m_{n,d} } )}

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data : RatioArray
        Fractionated data. :math:`r` in the equation above.
    ratio : RatioString
        The isotope ratio from which the fractionation factor should be calculated. :math:`n,d` in the equation above.
    std_abu : RatioArray, dict, optional
        An object with a ".get(item, default_value)" method containing the abundance of each isotope/ratio needed for
        this function. Default is the "best isotope fraction" reference values. :math:`R` in the equation above.
    std_mass : RatioArray, dict, optional
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values. :math:`m` in the equation above.

    Returns
    -------
    float
        The fractionation factor for *ratio* in *data*. :math:`\\alpha` in the equation above.
    """
    data = _e.check_type('data', data, _dt.RatioArray, coerce=True)
    ratio = _e.check_type('ratio', ratio, _dt.RatioString, coerce=True)
    std_abu = _e.check_reference_value('std_abu', std_abu, 'best isotope fraction')
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    return (_np.log(data.get(ratio) / std_abu.get(ratio)) / _np.log(std_mass.get(ratio)))


def remove_mass_fractionation(data, fractionation_factor, std_mass=None):
    """
    Remove exponential mass fractionation from the data.

    Calculated using:

    .. math::
        R_{n,d} = \\frac{r_{n,d}}{(m_{n,d})^{ \\alpha }}

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data : RatioArray
        Array containing data to be changed. :math:`r` in the equation above.
    fractionation_factor : float
        Fractionation factor to be applied. :math:`\\alpha` in the equation above
    std_mass : RatioArray, dict, optional
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values. :math:`m` in the equation above

    Returns
    -------
    RatioArray
        The corrected data. :math:`R` in the equation above.
    """
    data = _e.check_type('data', data, _dt.RatioArray, coerce=True)
    fractionation_factor = _e.check_type('fractionation_factor', fractionation_factor, _np.float, coerce=True)
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    return data / (std_mass.get(data.keys()) ** fractionation_factor)


def add_mass_fractionation(data, fractionation_factor, std_mass=None):
    """
    Add exponential mass fractionation to the data.

    Calculated using:

    .. math::
        R_{n,d} = r_{n,d} * (m_{n,d})^{ \\alpha }

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data : RatioArray
        Array containing data to be changed. :math:`r` in the equation above.
    fractionation_factor : float
        Fractionation factor to be applied. :math:`\\alpha` in the equation above
    std_mass : RatioArray, dict, optional
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values. :math:`m` in the equation above

    Returns
    -------
    RatioArray
        The corrected data. :math:`R` in the equation above.
    """
    data = _e.check_type('data', data, _dt.RatioArray, coerce=True)
    fractionation_factor = _e.check_type('fractionation_factor', fractionation_factor, _np.float, coerce=True)
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    return data * (std_mass.get(data.keys()) ** fractionation_factor)


def remove_isobaric_interferences(data, isotope, mf_factor = None, std_abu = None, std_mass=None):
    """
    Remove all isobaric interferences for a given element.

    Parameters
    ----------
    data : IsotopeArray, RatioArray
        The data that isobaric interferences should be removed from.
    isotope : IsotopeString
        The isotope of the interfering element that should be used for the correction. Must be present in *data*
    mf_factor : float, optional
        If given then this amount of exponential mass fractionation is added to the isobaric interferences values before
        they are removed from *data*.
    std_abu : IsotopeArray, dict
        An object with a ".get(item, default_value)" method containing the abundance of each isotope/ratio needed for
        this function.Default is the "best isotope fraction" reference values.
    std_mass : RatioArray, dict
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values.

    Returns
    -------
    IsotopeArray or RatioArray
        Data minus the isobaric interferences.
    """
    data = _e.check_type('data', data, _dt.RatioArray, coerce=True)
    isotope = _e.check_type('isotope', isotope, _dt.IsotopeString, coerce=True)
    mf_factor = _e.check_type('mf_factor', mf_factor, _np.float, coerce=True, allow_none=True)
    std_abu = _e.check_reference_value('std_abu', std_abu, 'best isotope fraction')
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    #Get the information we need to make the correction that works
    # for both isotope arrays and ratio arrays
    if isinstance(data, _dt.RatioArray):
        keys = data.keys()
        numer = keys.numerators()
        denom = keys.get_common_denominator()
    elif isinstance(data, _dt.IsotopeArray):
        keys = data.keys()
        numer = keys
        denom = None
    else:
        raise TypeError('variable "data" must be a IsotopeArray or a RatioArray not {}'.format(data.__class__.__name__))

    #Get the abundances of all isotopes of the interfering element
    inf_data = std_abu.get(std_abu.isotope_keys().copy(element_symbol=isotope.element_symbol))

    #Turn into a ratio relative to *isotope*
    inf_data = inf_data.ratio(isotope)

    #Account for mass fractionation of *mf_factor* is given
    if mf_factor is not None: inf_data = add_mass_fractionation(inf_data, mf_factor, std_mass=std_mass)

    #Scale relative to the measured value of *isotope*
    inf_data = inf_data * data[keys[numer.index(isotope)]]

    #Convert to a mass array for easy lookup later
    inf_data = _dt.MassArray(inf_data, keys=inf_data.keys().numerators().mass_numbers())

    #Create the output array
    out = _dt.asarray(data.nrows, keys=keys)

    #Loop through each key in *out* and remove interference
    for i in range(len(keys)):
        out[keys[i]] = data[keys[i]] - inf_data.get(numer[i].mass_number, 0)

    #If *data* is a ratio array then correct for any interference on the denominator
    if denom is not None:
        out = out / (1 - inf_data.get(denom.mass_number, 0))

    return out


def add_isobaric_interferences(data, isotope, value, mf_factor = None, std_abu=None, std_mass=None):
    """
    Add isobaric interferences of a given element to the data.

    Parameters
    ----------
    data : IsotopeArray, RatioArray
        The data that isobaric interferences should be added to.
    isotope : IsotopeString
        The isotope of the interfering element that should be used to add isobaric interferences. If not already
        present, a columns with this key will be added to the data.
    value : float
        This amount of *isotope* will be added to the data. The contribution from the other isotopes of the element
        is scaled to this value.
    mf_factor : float, optional
        If given then this amount of exponential mass fractionation is added to the isobaric interferences values before
        they are added to *data*.
    std_abu : IsotopeArray, dict
        An object with a ".get(item, default_value)" method containing the abundance of each isotope/ratio needed for
        this function.Default is the "best isotope fraction" reference values.
    std_mass : RatioArray, dict
        An object with a ".get(item, default_value)" method containing the masses of each isotope/ratio needed for
        this function. Default is the "isotope mass" reference values.

    Returns
    -------
    IsotopeArray or RatioArray
        Data plus the isobaric interferences.
    """
    data = _e.check_type('data', data, _dt.RatioArray, coerce=True)
    isotope = _e.check_type('isotope', isotope, _dt.IsotopeString, coerce=True)
    mf_factor = _e.check_type('mf_factor', mf_factor, _np.float, coerce=True, allow_none=True)
    std_abu = _e.check_reference_value('std_abu', std_abu, 'best isotope fraction')
    std_mass = _e.check_reference_value('std_mass', std_mass, 'isotope mass')

    # Get the information we need to make the correction that works
    # for both isotope arrays and ratio arrays
    if isinstance(data, _dt.RatioArray):
        keys = data.keys()
        numer = keys.numerators()
        denom = keys.get_common_denominator()
        out_keys = keys
        #Make sure that *isotope* is part of the returned array
        if isotope not in numer: out_keys.append(isotope/denom)
    elif isinstance(data, _dt.IsotopeArray):
        keys = data.keys()
        numer = keys
        denom = None
        out_keys = keys
        # Make sure that *isotope* is part of the returned array
        if isotope not in numer: out_keys.append(isotope)
    else:
        raise TypeError('variable "data" must be a IsotopeArray or a RatioArray not {}'.format(data.__class__.__name__))

    # Get the abundances of all isotopes of the interfering element
    inf_data = std_abu.get(std_abu.isotope_keys().copy(element_symbol=isotope.element_symbol))

    # Turn into a ratio relative to *isotope*
    inf_data = inf_data.ratio(isotope)

    # Account for mass fractionation of *mf_factor* is given
    if mf_factor is not None: inf_data = add_mass_fractionation(inf_data, mf_factor, std_mass=std_mass)

    # Scale relative to the measured value of *isotope*
    inf_data = inf_data * value

    # Convert to a mass array for easy lookup later
    inf_data = _dt.MassArray(inf_data, keys=inf_data.keys().numerators().mass_numbers())

    # Create the output array

    out = _dt.asarray(data.nrows, keys=out_keys)

    # Loop through each key in *out* and add interference
    for i in range(len(out_keys)):
        out[keys[i]] = data.get(keys[i], 0) + inf_data.get(numer[i].mass_number, 0)

    # If *data* is a ratio array then correct for any interference on the denominator
    if denom is not None:
        out = out / (1 + inf_data.get(denom.mass_number, 0))

    return out






import os as _os
import isopy.dtypes as _dtypes
from isopy import exceptions as _e
import csv as _csv
import datetime as _dt
import numpy as _np
import xlrd as _xlrd
import pkg_resources as _pkgr

__all__ = ['get_reference_values', 'import_exp', 'read_csv', 'write_csv', 'read_excel']

#This stores already loaded data sets to they dont have to be read from file each time they are requested.
reference_values = {}

def get_reference_values(name):
    """
    Get predefined reference values.

    The following reference values are currently avaliable:

    * "best isotope fraction M16" - Best measurement of isotopic abundances from a single terrestrial source. Taken
      from Table 1 in `Meija et al. (2016) Pure and Applied Chemistry, 88, 293-306
      <https://doi.org/10.1515/pac-2015-0503>`_.

    * "initial isotope fraction L09" - Isotope abundances 4.56 Ga ago. Taken from Table 10
      in `Lodders et al. (2009) Abundances of the Elements in Astronomy and Astrophysics, Landolt-Börnstein, Berlin
      <https://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

    * "initial isotope abundance L09" - Isotope abundances 4.56 Ga ago. Normalised to Si=10E6. Taken from Table 10
      in `Lodders et al. (2009) Abundances of the Elements in Astronomy and Astrophysics, Landolt-Börnstein, Berlin
      <https://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

    * "isotope mass W17" - The atomic mass of each isotope relative to :math:`^{12}C`. Taken from Table 1 in
      `Wang et al. (2017) Chinese Physics C, 41, 3 <https://doi.org/10.1088/1674-1137/41/3/030003>`_.

    * "sprocess isotope fraction B11" - The s-process contribution to each isotope as a fraction of the abundance.
      Taken from Table 5 in `Bisterzo et al. (2011) Monthly Notices of the
      Royal Astronomical Society, 418, 284-319 <https://doi.org/10.1111/j.1365-2966.2011.19484.x>`_.

    Functions that request default values will automatically receive the most recent data set if more than one exists.

    Parameters
    ----------
    name : str
        Name if the reference data set to be returned. If the last three reference characters (e.g. M16) are omitted
        then the most recent set of reference values with *name* will be returned. The data returned might therefore
        change in future versions. **It is therefore recommended that you specify the data set to be used.**

    Returns
    -------
    ReferenceDict
        A dict containing the specified reference values.
    """
    name = _e.check_type('name', name, str)
    name = name.strip()
    default_datasets = {'best isotope fraction': 'best isotope fraction M16',
                        'initial isotope fraction': 'initial isotope fraction L09',
                        'initial isotope abundance': 'initial isotope abundance L09',
                        'isotope mass': 'isotope mass W17',
                        'sprocess isotope fraction': 'sprocess isotope fraction B11'}

    #check if the reference characters were omitted and select the newst dataset.
    if name in default_datasets:
        name = default_datasets[name]

    #Check if the dataset has already been loaded
    if name in reference_values:
        return reference_values[name].copy()
    else:
        #filepath = _pkgr.resource_filename('isopy', 'referencedata/{}.csv'.format(name))
        filepath = _os.path.join(_os.path.dirname(__file__), 'referencedata', '{}.csv'.format(name))
        if not _os.path.exists(filepath):
            raise ValueError('A data set called "{}" does not exist'.format(name))

        #Load the reference values and add them to the
        reference_values[name] = _dtypes.ReferenceDict(read_csv(filepath))
        return reference_values[name].copy()


def import_exp(filename):
    """
    Return the data stored in the exported files from Neptune and Triton instruments.

    Parameters
    ----------
    filename : str
        Location of the **.exp** file to be read.

    Returns
    -------
    NeptuneData
        Stores the data from the exp file

    Examples
    --------
    Need to make examples
    """

    info = {}

    with open(filename, 'r') as exp:
        dialect = _csv.Sniffer().sniff(exp.read(1024))
        exp.seek(0)
        reader = _csv.reader(exp, dialect, quoting=_csv.QUOTE_NONE)

        #Check that we have right data
        title = reader.__next__()[0]
        if title != 'Neptune Analysis Data Report' and title != 'Triton Analysis Data Report' and False:
            raise TypeError('"{}" is not a valid neptune file'.format(filename))

        for row in reader:
            # Collect run information
            if ":" in row[0]:
                i = row[0].split(":", 1)
                info[i[0]] = i[1]

            # Cycle is the first field in the data table
            if row[0] == 'Cycle':
                cycle, time, isotope_data, ratio_data, other_data = _import_exp_data(row, reader)
                break

    return NeptuneData(info, cycle, time, isotope_data, ratio_data, other_data)


def _import_exp_data(legend_row, reader):
    idata = len(legend_row)
    data = [[] for i in range(idata)]

    for cycle in reader:
        # Signals the end of the cycle data
        if cycle[0] == "***" or cycle[1].strip() == '': break

        for i in range(idata):
            data[i].append(cycle[i])

    time = None
    cycles = None
    isotope_data = {}
    ratio_data = {}
    other_data = {}

    for i in range(idata):
        if legend_row[i] == 'Cycle':
            cycles = data[i]
        elif legend_row[i] == 'Time':
            time = data[i]
        else:
            if ":" in legend_row[i]:
                line, text = legend_row[i].split(":", 1)
                line = int(line)
            else:
                line = 1
                text = legend_row[i]

            if "/" in text:
                try: text = _dtypes.RatioString(text)
                except:
                    other_data.setdefault(line, {})[text] = data[i]
                else:
                    ratio_data.setdefault(line, {})[text] = data[i]
            else:
                try: text = _dtypes.IsotopeString(text)
                except:
                    other_data.setdefault(line, {})[text] = data[i]
                else:
                    isotope_data.setdefault(line, {})[text] = data[i]

    return cycles, time, isotope_data, ratio_data, other_data


class NeptuneData():
    """
    Returned by :func:`read_exp`.

    If no line numbers are given in the exp file then the line number is set to 1.

    Attributes
    ----------
    info : dict
        The information stored in the exp file before the isotope data.
    cycle : np.ndarray
        An array containing the cycle number of each row in the data.
    isotope_data : dict
        A dict of :class:`IsotopeArray` each containing the isotope data for the line given by the dict key.
    ratio_data : dict
        A dict of :class:`RatioArray` each containing the ratio data for the line given by the dict key.
    other_data : dict
        A dict of dict each containing the remaining data for the line given by the dict key.
    """
    def __init__(self, information, cycle, time, isotope_data, ratio_data, other_data):
        self.info = information.copy()
        self.cycle = _np.array(cycle, dtype = 'int')
        self.time = [_dt.datetime.strptime(time[i], '%H:%M:%S:%f') for i in range(len(time))]

        self.isotope_data = isotope_data
        self.ratio_data = ratio_data
        self.other_data = other_data


def read_csv(filepath, comment_symbol='#', skip_columns=None):
    """
    Read data from an csv file.

    The first row of the file will be used as the key and the subsequent rows will be taken as the value.

    Parameters
    ----------
    filepath : str
        Location of file to be read
    comment_symbol : str, optional
        Rows at the beginning of the file that start with this *comment_symbol* are ignored. Default value is ``"#"``.
    skip_columns : int, list of int, optional
        If given then the columns with these indexes are not read.

    Returns
    -------
    dict
        A dict containing the contents of the file.

    Examples
    --------
    Need to make examples
    """
    filepath = _e.check_type('filepath', filepath, str)
    comment_symbol = _e.check_type('comment_symbol', comment_symbol, str)
    skip_columns = _e.check_type_list('skip_columns', skip_columns, int, allow_none=True, make_list=True
                                      )
    keys = []
    values = []
    key_len = None

    if skip_columns is None: skip_columns = []
    elif not isinstance(skip_columns, (list, tuple)): skip_columns = [skip_columns]

    with open(filepath, 'r', newline='') as file:
        firstline = file.readline()
        file.seek(0)
        dialect = _csv.Sniffer().sniff(firstline)
        if firstline[:3] == 'ï»¿':
            #Ugly way of detecting if file is encoded with a Byte order mark
            #Excel csv files are encoded like this
            file.seek(3)
        reader = _csv.reader(file, dialect=dialect)

        for row in reader:
            if [len(r.strip()) for r in row].count(0) == len(row): break
            if comment_symbol is None or not row[0].strip().startswith(comment_symbol):
                if key_len is None:
                    keys = [x.strip() for x in row]
                    while keys[-1] == '': keys = keys[:-1]
                    key_len = len(keys)
                    values = [list([]) for x in keys]
                else:
                    if len(row) < key_len: raise ValueError('Row does not have a value for every key')
                    for i in range(key_len):
                        values[i].append(row[i].strip())
    return {keys[i]: values[i] for i in range(key_len) if i not in skip_columns}


def write_csv(data, filepath, comments = None, delimiter =',', comment_symbol='#'):
    """
    Write data to a csv file.

    The first row in each column will contain the key of each item in *data*. Subsequent rows will contains the
    values of each item in the dict.

    Parameters
    ----------
    data : dict, IsopyArray
        The data to be writen to the file.
    filepath : str
        Location where to save the file. It the file already exists it will overwrite the old file.
    comments : str, list of str, optional
        Comments to be included before the file.
    delimiter :str, optional
        The delimiter for columns in the file. Default value is ``","``
    comment_symbol : str, optional
        Used to denote that the line contains a comment.  Default value is ``"#"``.
    """
    filepath = _e.check_type('filepath', filepath, str)
    comments = _e.check_type_list('comments', comments, str, allow_none=True, make_list=True)
    delimiter = _e.check_type('delimiter', delimiter, str)
    comment_symbol = _e.check_type('comment_symbol', comment_symbol, str)

    with open(filepath, mode='w', newline='') as file:
        writer = _csv.writer(file, delimiter=delimiter)

        if comments is not None:
            for c in comments:
                writer.writerow(['{}{}'.format(comment_symbol, c)])

        writer.writerow([k for k in data.keys()])

        if data.ndim == 0:
            writer.writerow(['{}'.format(data[k]) for k in data.keys()])
        else:
            keys = data.keys()
            for i in range(data.size):
                writer.writerow(['{}'.format(data[k][i]) for k in keys])


def read_excel(filename, sheetname = None, comment_symbol='#'):
    """
    Read data from sheets in an excel file.

    The first row of the sheet will be used as the key and the subsequent rows will be taken as the value.

    Parameters
    ----------
    filename : str
        Location of the file to be opened.
    sheetname : int, str, None, optional
        If ``int`` then the sheet at this index will be returned. If ``str`` then the sheet with this name will be
        returned. If ``None`` a dict containing the data for all the sheets in the array will be returned.
    comment_symbol : str, optional
        Rows at the top of each sheet starting with *comment_symbol* will be ignored. Default value is ``"#"``.

    Returns
    -------
    dict
        If *sheetname* is ``None`` a dict of each sheet in the file will be returned. Otherwise a dict of the data in
        the given sheet is returned.

    Examples
    --------
    Need to make examples
    """
    wb = _xlrd.open_workbook(filename)
    try:
        if sheetname is None:
            return {sheetname: _read_sheet(wb.sheet_by_name(sheetname), comment_symbol) for sheetname in wb.sheet_names()}
        elif isinstance(sheetname, int):
            return _read_sheet(wb.sheet_by_index(sheetname), comment_symbol)
        else:
            return _read_sheet(wb.sheet_by_name(sheetname), comment_symbol)
    finally:
        wb.release_resources()


def _read_sheet(sheet, comment_symbol):
    for ri in range(sheet.nrows):
        if sheet.cell_value(ri, 0)[:len(comment_symbol)] != comment_symbol:
            return {sheet.cell_value(ri, c): [sheet.cell_value(r, c) for r in (range(ri+1, sheet.nrows))] for c in
                    range(sheet.ncols)}


def write_excel(filename, *sheets):
    pass


from matplotlib.axes import Axes as _Axes
from matplotlib.patches import Polygon as _Polygon
import numpy as _np
from isopy import exceptions as _e
from isopy import dtypes as _dt



__add__ = ['plot_york_regression', 'york_regression', 'johnson_nyquist_noise']


def plot_york_regression(axes, regression, color = 'black', slopeline = True, errorline = True, shaded = True,
                         zorder = 1, shade_kwargs = None, **line_kwargs):
    """
    Plots the regression within the limits of the supplies axes.

    Parameters
    ----------
    axes : matplotlib.axes.Axes, matplotlib.pytplot.plt
        An axes instance where the regression envelope will be plotted.
    regression : YorkRegression
        The regression to be plotted
    color : str
        The color of the lines/polygon drawn for this regression. See
        `matplotlib documentation <https://matplotlib.org/tutorials/colors/colors.html>`_.
    slopeline : bool
        If ``True`` the slope will be plotted as a solid line. Default value is ``True``.
    errorline : bool, optional
        If ``True`` the outline of the regression envelope will be plotted as a dashed line. Default value is ``True``.
    shaded : bool, optional
        If ``True`` the regression envelope will be shaded. Default value is ``True``.
    zorder : int, optional
        The higher the value the later the regression will be be drawn. Default value is ``1``.
    shade_kwargs : dict, optional
        kwargs that will be passed to polygon that makes up the shaded area.
    **line_kwargs
        Keyword arguments passed when drawing the slopeline and the errorline.

    """
    #TODO add
    slopeline_kwargs = {'linestyle': '-', 'marker': '', 'color': color, 'zorder': zorder}
    slopeline_kwargs.update(line_kwargs)
    errorline_kwargs = {'linestyle': '--', 'marker': '', 'color': color, 'zorder': zorder}
    errorline_kwargs.update(line_kwargs)
    errorline_kwargs.pop('label', None)
    poly_kwargs = {'color': color, 'alpha': 0.1, 'zorder': zorder-0.1}
    if shade_kwargs is not None: poly_kwargs.update(shade_kwargs)
    if not isinstance(axes, _Axes):
        try:
            axes = axes.gca()
        except:
            raise ValueError('axes must be a matplotlib Axes or pyplot obj')

    xlim = axes.get_xlim()
    ylim = axes.get_ylim()

    xval = _np.linspace(xlim[0], xlim[1], 10000)
    slope = xval * regression.slope + regression.intercept
    upper = slope + regression.y_error(xval)
    lower = slope - regression.y_error(xval)

    index_slope = _np.all([slope > ylim[0], slope < ylim[1]], 0)
    index_lower = _np.all([lower > ylim[0], lower < ylim[1]], 0)
    index_upper = _np.all([upper > ylim[0], upper < ylim[1]], 0)

    upper_coordinates = [*zip(xval[index_upper], upper[index_upper])]
    lower_coordinates = [*zip(xval[index_lower], lower[index_lower])]
    lower_coordinates.reverse()

    if shaded:
        poly = _Polygon(upper_coordinates + lower_coordinates, **poly_kwargs)
        axes.add_patch(poly)

    if slopeline: axes.plot(xval[index_slope], slope[index_slope], **slopeline_kwargs)
    if errorline:
        axes.plot(xval[index_upper], upper[index_upper], **errorline_kwargs)
        axes.plot(xval[index_lower], lower[index_lower], **errorline_kwargs)

    #axes.show(xlim, [regression.slope*xlim[0]+regression.intercept, regression.slope*xlim[1]+regression.intercept], **kwargs)


def york_regression(X, Y, Xerr, Yerr, r=0, tol=0.00001):
    """
    Calculate the slope and intercept taking into account the uncertainty on x and y.

    Uncertainties on the slope and intercept are given as 2σ. Based on the formulas from `York et al. (2004)
    American Journal of Physics 72, 367 <https://doi.org/10.1119/1.1632486>`_.

    Parameters
    ----------
    X : np.ndarray
        Values for x axis
    Y : np.ndarray
        Values for y axis
    Xerr : float, np.ndarray
        Uncertainties for x axis. If a single value then it will be used for all values of *X*.
    Yerr : float, np.ndarray
        Uncertainties for y axis. If a single value then it will be used for all values of *Y*.
    r : float, optional
        Correlation coefficient between errors in x and y. Default value is ``0``.
    tol : float, optional
        Tolerance for fit. Default value is ``0.00001``.


    Returns
    -------
    YorkRegression
        Contains the slope and intercept values and all information needed to calculate the y for any given x.

    See Also
    --------
    :class:`YorkRegression`, :func:`plot_york_regression`
    """
    X = _e.check_type('X', X, _np.ndarray, coerce=True, coerce_into=_np.array)
    Y = _e.check_type('Y', Y, _np.ndarray, coerce=True, coerce_into=_np.array)
    Xerr = _e.check_type('Xerr', Xerr, _np.ndarray, _np.float, coerce=True, coerce_into=[_np.float, _np.array])
    Yerr = _e.check_type('Yerr', Yerr, _np.ndarray, _np.float, coerce=True, coerce_into=[_np.float, _np.array])
    r = _e.check_type('r', r, _np.float, coerce=True)
    tol = _e.check_type('tol', tol, _np.float, coerce=True)

    if r > 1 or r < 0:
        raise ValueError('r must be between 0 and 1')
    if X.ndim != 1:
        raise ValueError('parameter "X": must have 1 dimension not {}'.format(X.ndim))
    if Y.ndim != 1:
        raise ValueError('parameter "Y": must have 1 dimension not {}'.format(X.ndim))

    if X.size != Y.size:
        raise ValueError('X and Y must have the same size')

    if not isinstance(Xerr, _np.ndarray):
        Xerr = _np.ones(X.size) * Xerr
    elif Xerr.size != X.size:
        raise ValueError('Xerr and X must have the same size')

    if not isinstance(Yerr, _np.ndarray):
        Yerr = _np.ones(Y.size) * Yerr
    elif Yerr.size != Y.size:
        raise ValueError('Yerr and Y must have the same size')

    # Do calculations
    WX = 1 / (Xerr ** 2)
    WY = 1 / (Yerr ** 2)

    k = _np.sqrt(WX * WY)
    slope = (_np.mean(X * Y) - _np.mean(X) * _np.mean(Y)) / (_np.mean(X ** 2) - _np.mean(X) ** 2)

    # Iterativley converge self.slope until tol is reached
    for i in range(1000):
        W = (WX * WY) / (WX + (slope ** 2) * WY - 2 * slope * r * k)
        Xbar = _np.sum(W * X) / _np.sum(W)
        Ybar = _np.sum(W * Y) / _np.sum(W)
        U = X - Xbar
        V = Y - Ybar
        beta = W * (U / WY + (slope * V) / WX - (slope * U + V) * (r / k))

        b2 = _np.sum(W * beta * V) / _np.sum(W * beta * U)
        dif = _np.sqrt((b2 / slope - 1) ** 2)
        slope = b2
        if dif < tol: break
    if i == 999:
        raise ValueError('Unable to calculate a fit after 1000 iterations.')

    # Calculate self.intercept
    intercept = Ybar - slope * Xbar

    # Calculate adjusted points
    x = Xbar + beta
    xbar = _np.sum(W * x) / _np.sum(W)
    u = x - xbar

    y = Ybar + slope * beta
    ybar = _np.sum(W * y) / _np.sum(W)
    v = y - ybar

    #Calculate the uncertianty on the slope
    slope_error = _np.sqrt(1 / _np.sum(W * (u ** 2)))
    intercept_error = _np.sqrt(1 / _np.sum(W) + ((0 - xbar) ** 2) * (slope_error ** 2))

    #Goodness of fit
    S = _np.sum(W * (Y - slope * X - intercept) ** 2)

    #implementation in isoplotr
    return YorkRegression(slope, slope_error, intercept, intercept_error, _np.sum(W), xbar)


class YorkRegression:
    """
    Returned by :func:`york_regression`.

    This object contains all the necessary attributes to calculate the value and uncertainty of y for any given
    value of x.

    Attributes
    ----------
    slope : float
        The best fit slope of the regression
    slope_error : float
        The uncertainty of the slope
    intercept : float
        The y axis intercept of the regression
    intercept_error : float
        The uncertainty of the y axis intercept
    sumW : float
        Sum of the weights
    xbar : float
        The weighted x mean.

    See Also
    --------
    :func:`york_regression`, :func:`plot_york_regression`
    """
    def __init__(self, slope, slope_error, intercept, intercept_error, sumW, xbar):
        self.slope = slope
        self.slope_error = slope_error
        self.intercept = intercept
        self.intercept_error = intercept_error
        self.sumW = sumW
        self.xbar = xbar

    def __repr__(self):
        return 'y=({} +- {})x + ({} +- {}); xbar = {}, sum(W) = {}'.format(self.slope, self.slope_error,
                                                self.intercept, self.intercept_error, self.xbar, self.sumW)

    def y_error(self, x):
        """
        Return the uncertainty of y at *x*.
        """
        return _np.sqrt(1 / self.sumW + ((x - self.xbar) ** 2) * (self.slope_error ** 2))

    def y(self, x):
        """
        Return y at *x*.
        """
        return self.slope * x + self.intercept


def johnson_nyquist_noise(voltage, resistor = 1E11, integration_time = 8.389, include_counting_statistics = True,
                          T = 309, R = 1E11, cpv = 6.25E7):
    """
    Calculate the Johnson-Nyquist noise and counting statistics for a given voltage.

    The Johnson-Nyquist noise (:math:`n_{jn}` is calculated as:

    .. math::
        n_{jn} = \\sqrt{ \\frac{4*k_{b}*T*r} {t} } * \\frac{R} {r}

    The counting statistics, or shot noise, (:math:`n_{cs}` is calculated as:

    .. math::
        n_{cs} = \\sqrt{ \\frac{1} {v * c_{V} * t}} * v

    The two are combined as:

    .. math::
        n_{all} = \\sqrt{ (n_{jn})^2 + (n_{cs})^2 }

    The noise for a specific isotope ratio can be calculated as:

    .. math::
        n_{n,d} = \\sqrt{ (n_{n})^2 + (n_{d})^2 }

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Adapted from the equations in `Liu & Pearson (2014) Chemical Geology, 10, 301-311
    <https://doi.org/10.1016/j.chemgeo.2013.11.008>`_.

    Parameters
    ----------
    voltage : IsotopeArray, float, np.ndarray
        The measured voltages. :math:`v` in the equations above.
    resistor : IsotopeArray, float, np.ndarray
        The resistor for the measurement. Default value is ``1E11``. :math:`r` in the equations above.
    integration_time : float, optional
        The integration time in seconds for a single measurement. Default value is ``8.389``.
        :math:`t` in the equations above.
    include_counting_statistics: bool, optional
        If ``True`` then the counting statistics are included in the returned value. Default value is ``True``.
    T : float, optional
        Amplifier housing temperature in kelvin. Default value is ``309``. :math:`T` in the equations above.
    R : float, optional
        Voltage is reported as volts for this resistor value. Default value is ``1E11``. :math:`R` in the equations
        above.
    cpv : float, optional
        Counts per volt per second. Default value is ``6.25E7``. :math:`C_{V}` in the equations above.

    Returns
    -------
    np.float or np.ndarray or IsotopeArray
        The noise in V for the given voltage/set of voltages.
    """

    voltage = _e.check_type('voltage', voltage, _dt.IsotopeArray, _np.ndarray, _np.float, coerce=True,
                            coerce_into=[_dt.IsotopeArray, _np.float, _np.array])
    resistor = _e.check_type('resistor', resistor, _dt.IsotopeArray, _np.ndarray, _np.float, coerce=True,
                             coerce_into=[_dt.IsotopeArray, _np.float, _np.array])
    integration_time = _e.check_type('integration_time', integration_time, _np.float, coerce=True)
    include_counting_statistics = _e.check_type('include_counting_statistics', include_counting_statistics,
                                                bool)
    T = _e.check_type('T', T, _np.float, coerce=True)
    R = _e.check_type('R', R, _np.float, coerce=True)
    cpv = _e.check_type('cpv', cpv, _np.float, coerce=True)

    kB = _np.float(1.3806488E-023) # Boltsman constant

    t_noise = _np.sqrt((4 * kB * T * resistor) / integration_time) / (voltage * (resistor / R)) * voltage
    t_noise = _np.sqrt((4 * kB * T * resistor) / integration_time) * (R / resistor)

    if include_counting_statistics:
        c_stat = _np.sqrt(1 / (voltage * cpv * integration_time)) * voltage
        return _np.sqrt(c_stat ** 2 + t_noise ** 2)
    else:
        return t_noise





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="isopy",
    version="0.3.1",
    packages=setuptools.find_packages(include=['isopy', 'isopy.*']),
    install_requires=['numpy', 'tables', 'pyperclip', 'xlrd', 'matplotlib'],

    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

    package_data = {
        "": ["*.txt", "*.rst", "*.csv", "*.xlsx"]},

    author="Mattias Ek",
    author_email="mattias.ek@bristol.ac.uk",
    description="A python package for data processing in geo/cosmochemistry",
    long_description=long_description,
    keywords="array isotope geochemistry cosmochemisty",
    long_description_content_type="text/markdown",
    url="https://github.com/mattias-ek/isopy",
    download_url = "https://github.com/mattias-ek/isopy/archive/v0.3.1.tar.gz",

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha"
    ],
)